import pennylane as qml
from datetime import datetime
import time


def gradient_descent(
    cost,
    init_settings,
    num_steps=150,
    step_size=0.1,
    sample_width=25,
    grad_fn=None,
    verbose=True,
    interface="autograd",
):
    """Performs a numerical gradient descent optimization on the provided ``cost`` function.
    The optimization is seeded with (random) ``init_settings`` which are then varied to
    minimze the cost.

    :param cost: The cost function to be minimized with gradient descent.
    :type cost: function

    :param init_settings: A valid input for the cost function.
    :type init_settings: array-like[float]

    :param num_steps: The number of gradient descent iterations, defaults to ``150``.
    :type num_steps: int, optional

    :param step_size: The learning rate for the gradient descent, defaults to ``0.1``.
    :type step_size: float, optional

    :param sample_width: The number of steps between "sampled" costs which are printed/returned
        to user, defaults to ``25``.
    :type sample_width: int, optional

    :param grad_fn: A custom gradient function, default to ``None`` which applies the standard numerical gradient.
    :type grad_fn: function, optional

    :param verbose: If ``True``, progress is printed during the optimization, defaults to ``True``.
    :type verbose: bool, optional

    :param interface: Specifies the optimizer software either ``"autograd"`` or ``"tf"`` (TensorFlow).
    :type interface: string, default ``"autograd``"

    :return: Data regarding the gradient descent optimization.
    :rtype: dictionary, contains the following keys:

        * **opt_score** (*float*) - The maximized reward ``-(min_cost)``.
        * **opt_settings** (*array-like[float]*) - The setting for which the optimum is achieved.
        * **scores** (*array[float]*) - The list of rewards sampled during the gradient descent.
        * **samples** (*array[int]*) - A list containing the iteration for each sample.
        * **settings_history** (*array[array-like]*) - A list of all settings found for each
          intermediate step of gradient descent
        * **datetime** (*string*) - The date and time in UTC when the optimization occurred.
        * **step_times** (*list[float]*) - The time elapsed during each sampled optimization step.
        * **step_size** (*float*) - The learning rate of the optimization.

    .. warning::

        The ``gradient_descent`` function minimizes the cost function, however, the general
        use case within this project is to maximize the violation of a Bell inequality.
        The maximization is assumed within ``gradient_descent`` and is applied by multiplying
        the cost by (-1). This is an abuse of function naming and will be resolved in a future
        commit by having ``gradient_descent`` return the minimized cost rather than the maximized
        reward. The resolution is to wrap ``gradient_descent`` with a ``gradient_ascent`` function
        which maximizes a reward function equivalent to ``-(cost)``.

    :raises ValueError: If the ``interface`` is not supported.
    """

    if interface == "autograd":
        opt = qml.GradientDescentOptimizer(stepsize=step_size)
    elif interface == "tf":
        from .lazy_tensorflow_import import tensorflow as tf

        opt = tf.keras.optimizers.SGD(learning_rate=step_size)
    else:
        raise ValueError('Interface "' + interface + '" is not supported.')

    settings = init_settings
    scores = []
    samples = []
    step_times = []
    settings_history = [init_settings]

    start_datetime = datetime.utcnow()
    elapsed = 0

    # performing gradient descent
    for i in range(num_steps):
        if i % sample_width == 0:
            score = -(cost(settings))
            scores.append(score)
            samples.append(i)

            if verbose:
                print("iteration : ", i, ", score : ", score)

        start = time.time()
        if interface == "autograd":
            settings = opt.step(cost, settings, grad_fn=grad_fn)
        elif interface == "tf":
            # opt.minimize updates settings in place
            tf_cost = lambda: cost(settings)
            opt.minimize(tf_cost, settings)

        elapsed = time.time() - start

        if i % sample_width == 0:
            step_times.append(elapsed)

            if verbose:
                print("elapsed time : ", elapsed)

        settings_history.append(settings)

    opt_score = -(cost(settings))
    step_times.append(elapsed)

    scores.append(opt_score)
    samples.append(num_steps)

    return {
        "datetime": start_datetime.strftime("%Y-%m-%dT%H:%M:%SZ"),
        "opt_score": opt_score,
        "opt_settings": settings,
        "scores": scores,
        "samples": samples,
        "settings_history": settings_history,
        "step_times": step_times,
        "step_size": step_size,
    }



import pennylane as qml
from pennylane import numpy as np
from pennylane import math


class NoiseNode:
    """A class that configures each noise node in the quantum network.

    :param wires: A list of wires on which the node is defined.
    :type wires: array[int]

    :param quantum_fn: A PennyLane quantum function which accepts as input the
        positional arguments ``(settings, wires)`` where settings is an *array[float]*
        of length ``num_settings``.
    :type quantum_fn: function

    :returns: An instantiated ``NoiseNode`` class.
    """

    def __init__(self, wires, quantum_fn):
        self.wires = wires
        self.ansatz_fn = quantum_fn
        self.num_settings = 0


class PrepareNode(NoiseNode):
    """A class that configures each preparation node in the quantum network.

    :param num_in: The number of classical inputs for the node.
    :type num_in: int

    :param wires: A list of wires on which the node is defined.
    :type wires: array[int]

    :param quantum_fn: A PennyLane quantum function which accepts as input the
        positional arguments ``(settings, wires)`` where settings is an *array[float]*
        of length ``num_settings``.
    :type quantum_fn: function

    :param num_settings: The number of settings that the quantum function accepts.
    :type num_settings: int

    :param static_settings: A set of fixed settings for the preparation node.
    :type static_settings: optional, array-like

    :returns: An instantiated ``PrepareNode`` class.
    """

    def __init__(self, num_in, wires, quantum_fn, num_settings, static_settings=[]):
        super().__init__(wires, quantum_fn)
        self.num_in = num_in
        self.num_settings = num_settings
        self.settings_dims = (num_in, num_settings)
        self.static_settings = static_settings


class MeasureNode(PrepareNode):
    """A class that configures each measurement node in the quantum network.

    :param num_in: The number of classical inputs for the node.
    :type num_in: int

    :param num_out: The number of classical outputs for the node.
    :type num_out: int

    :param wires: A list of wires on which the node is defined.
    :type wires: array[int]

    :param quantum_fn: A PennyLane quantum function which accepts as input the
        positional arguments ``(settings, wires)`` where settings is an *array[float]*
        of length ``num_settings``.
    :type quantum_fn: function

    :param num_settings: The number of settings that the quantum function accepts.
    :type num_settings: int

    :param static_settings: A set of fixed settings for the preparation node.
    :type static_settings: optional, array-like

    :returns: An instantiated ``MeasureNode`` class.
    """

    def __init__(self, num_in, num_out, wires, quantum_fn, num_settings, static_settings=[]):
        super().__init__(num_in, wires, quantum_fn, num_settings, static_settings=static_settings)
        self.num_out = num_out


class NetworkAnsatz:
    """The ``NetworkAnsatz`` class describes a parameterized quantum prepare and measure network.
    The ansatz is constructed from a prepare layer and a measure layer.
    The prepare layer is a collection of unitaries which prepare quantum states while the
    measure layer is a collection of unitaries which encode the measurement basis.
    These layers are constructed from the ``prepare_nodes`` and ``measure_nodes`` inputs respectively.

    :param prepare_nodes: A list of ``PrepareNode`` classes.
    :type prepare_nodes: list[PrepareNode]

    :param measure_nodes: A list of ``MeasureNode`` classes.
    :type measure_nodes: list[MeasureNode]

    :param noise_nodes: A list of ``NoiseNode`` classes.
    :type noise_nodes: *optional* list[NoiseNode]

    :param dev_kwargs: Keyword arguments for the `pennylane.device`_ function.
    :type dev_kwargs: *optional* dict

    .. _pennylane.device: https://pennylane.readthedocs.io/en/stable/code/api/pennylane.device.html?highlight=device#qml-device

    :returns: An instantiated ``NetworkAnsatz`` class with the following fields:

    * **prepare_nodes** - The list of ``PrepareNode`` classes.
    * **measure_nodes** - The list of ``MeasureNode`` classes.
    * **prepare_wires** - The list of wires used by the ``prepare_nodes``.
    * **measure_wires** - The list of wires used by the ``measure_nodes``.
    * **network_wires** - The list of wires used by the network ansatz.
    * **dev_kwargs** - *mutable*, the keyword args to pass to the `pennylane.device`_ function.
                       If no ``dev_kwargs`` are provided, a ``"default.qubit"`` is constructed for
                       noiseless networks and ``"default.mixed"`` device is constructed
                       for noisy networks.
    * **dev** (*qml.device*) - *mutable*, the most recently constructed device for the ansatz.
    * **fn** (*function*) - A quantum function implementing the quantum network ansatz.

    :raises ValueError: If the wires for each ``PrepareNode`` (or ``MeasureNode``) are not unique.
    """

    def __init__(self, prepare_nodes, measure_nodes, noise_nodes=[], dev_kwargs=None):
        self.prepare_nodes = prepare_nodes
        self.measure_nodes = measure_nodes
        self.noise_nodes = noise_nodes

        self.prepare_wires = self.collect_wires(prepare_nodes)
        self.measure_wires = self.collect_wires(measure_nodes)
        self.noise_wires = (
            qml.wires.Wires([]) if self.noise_nodes == [] else self.collect_wires(noise_nodes)
        )

        self.network_wires = qml.wires.Wires.all_wires(
            [self.prepare_wires, self.measure_wires, self.noise_wires]
        )

        default_dev_name = "default.qubit" if len(self.noise_nodes) == 0 else "default.mixed"
        self.dev_kwargs = dev_kwargs or {"name": default_dev_name}
        self.dev_kwargs["wires"] = self.network_wires
        self.dev = self.device()

        self.fn = self.construct_ansatz_circuit()

    def set_device(self, name, **kwargs):
        """Configures a new PennyLane device for executing the network ansatz circuit.
        For more details on parameters see `pennylane.device`_.
        This method updates the values stored in ``self.dev_kwargs`` and ``self.dev``.

        :returns: The instantiated device.
        :rtype: ``pennylane.device``
        """

        dev_kwargs = kwargs.copy() if kwargs else {}
        dev_kwargs["name"] = name
        dev_kwargs["wires"] = self.network_wires

        self.dev_kwargs = dev_kwargs
        self.dev = qml.device(**self.dev_kwargs)
        return self.dev

    def device(self):
        """Instantiates a new PennyLane device configured using the ``self.dev_kwargs`` parameters.
        A distinct device is created each time this function runs.

        :returns: The instantiated device.
        :rtype: ``pennylane.device``
        """
        self.dev = qml.device(**self.dev_kwargs)
        return self.dev

    def construct_ansatz_circuit(self):
        prepare_layer = self.circuit_layer(self.prepare_nodes)
        noise_layer = self.circuit_layer(self.noise_nodes)
        measure_layer = self.circuit_layer(self.measure_nodes)

        noise_settings = [np.array([]) for i in range(len(self.noise_nodes))]

        num_prep_settings = math.sum([node.num_settings for node in self.prepare_nodes])

        def ansatz_circuit(settings):
            prep_settings = settings[0:num_prep_settings]
            meas_settings = settings[num_prep_settings:]

            prepare_layer(prep_settings)
            noise_layer(noise_settings)
            measure_layer(meas_settings)

        return ansatz_circuit

    @staticmethod
    def collect_wires(network_nodes):
        """A helper method for the ``NetworkAnsatz`` class which collects and aggregates the
        wires from a set of collection of network nodes (``prepare_nodes`` or ``measure_nodes``).

        :param network_nodes: A list consisting of either ``PrepareNode``'s or ``MeasureNode``'s.
        :type network_nodes: list[PrepareNode or MeasureNode]

        :raises ValueError: If the same wire is used in two different nodes in ``network_nodes``.
        """
        ansatz_wires = list(map(lambda node: qml.wires.Wires(node.wires), network_nodes))
        all_wires = qml.wires.Wires.all_wires(ansatz_wires)
        unique_wires = qml.wires.Wires.unique_wires(ansatz_wires)

        # two nodes cannot prepare a state on the same wire
        if not all_wires.tolist() == unique_wires.tolist():
            raise ValueError(
                "One or more wires are not unique. Each node must contain unique wires."
            )

        return all_wires

    @staticmethod
    def layer_settings(scenario_settings, node_inputs, nodes):
        """Constructs the list of settings for a circuit layer in the network ansatz.

        The type of tensor used for the returned layer settings matches the tensor
        type of the elements of ``scenario_settings``.

        :param scenario_settings: A list containing the settings for all classical inputs.
        :type network_nodes: list[array[float]]

        :param node_inputs: A list of the classical inputs supplied to each network node.
        :type node_inputs: list[int]

        :returns: A 1D array of all settings for the circuit layer.
        :rtype: array[float]
        """
        return math.concatenate(
            [
                scenario_settings[i][node_input]
                if len(nodes[i].static_settings) == 0
                else nodes[i].static_settings[node_input]
                for i, node_input in enumerate(node_inputs)
            ]
        )

    def qnode_settings(self, scenario_settings, prep_inputs, meas_inputs):
        """Constructs a list of settings to pass to the qnode executing the network ansatz.

        :param scenario_settings: The settings for the network ansatz scenario.
        :type scenario_settings: list[list[np.ndarray]]

        :param prep_inputs: The classical inputs passed to each preparation node.
        :type prep_inputs: list[int]

        :param meas_inputs: The classical inputs passed to each measurement node.
        :type meas_inputs: list[int]

        :returns: The settings to pass to the constructed qnode.
        :rtype: list[float]
        """

        prep_settings = self.layer_settings(scenario_settings[0], prep_inputs, self.prepare_nodes)
        meas_settings = self.layer_settings(scenario_settings[1], meas_inputs, self.measure_nodes)
        return math.concatenate([prep_settings, meas_settings])

    @staticmethod
    def circuit_layer(network_nodes):
        """Constructs a quantum function for an ansatz layer of provided network nodes.

        :param network_nodes: A list of network nodes which can be either
                              ``NoiseNode``, ``PrepareNode``, or ``MeasureNode``
        :type network_nodes: list[NetworkNode]

        :returns: A quantum function evaluated as ``circuit(settings)`` where ``settings``
                  is an array constructed via the ``layer_settings`` function.
        :rtype: function
        """

        def circuit(settings):
            current_id = 0
            for node_id in range(len(network_nodes)):
                node = network_nodes[node_id]
                node_settings = settings[current_id : current_id + node.num_settings]

                current_id += node.num_settings

                node.ansatz_fn(node_settings, node.wires)

        return circuit

    def rand_scenario_settings(self):
        """Creates a randomized settings array for the network ansatz.

        :returns: A nested list containing settings for each network node.
                  ``PreparNode`` settings are listed under index ``0`` while
                  ``MeasureNode`` settings are listed under index ``1``.
                  The measure and prepare layers settings are a list of numpy arrays
                  where the dimension of each array is ``(num_inputs, num_settings)``.
        :rtype: list[list[np.array]]
        """
        prepare_settings = [
            2 * np.pi * np.random.random(node.settings_dims) - np.pi
            if len(node.static_settings) == 0
            else np.array([[]])
            for node in self.prepare_nodes
        ]
        measure_settings = [
            2 * np.pi * np.random.random(node.settings_dims) - np.pi
            if len(node.static_settings) == 0
            else np.array([[]])
            for node in self.measure_nodes
        ]

        return [prepare_settings, measure_settings]

    def tf_rand_scenario_settings(self):
        """Creates a randomized settings array for the network ansatz using TensorFlow
        tensor types.

        :returns: See :meth:`qnetvo.NetworkAnsatz.rand_scenario_settings` for details.
        :rtype: list[list[tf.Tensor]]
        """
        from .lazy_tensorflow_import import tensorflow as tf

        np_settings = self.rand_scenario_settings()

        return [
            [tf.Variable(settings) for settings in np_settings[0]],
            [tf.Variable(settings) for settings in np_settings[1]],
        ]

    def zero_scenario_settings(self):
        """Creates a settings array for the network ansatz that consists of zeros.

        :returns: A nested list containing settings for each network node.
                  ``PrepareNode`` settings are listed under index ``0`` while
                  ``MeasureNode`` settings are listed under index ``1``.
                  The measure and prepare layers settings are a list of numpy arrays
                  where the dimension of each array is ``(num_inputs, num_settings)``.
        :rtype: list[list[np.array]]
        """
        prepare_settings = [
            np.zeros(node.settings_dims) if len(node.static_settings) == 0 else np.array([[]])
            for node in self.prepare_nodes
        ]
        measure_settings = [
            np.zeros(node.settings_dims) if len(node.static_settings) == 0 else np.array([[]])
            for node in self.measure_nodes
        ]

        return [prepare_settings, measure_settings]



import pennylane as qml
import pennylane.numpy as np
from pennylane import math
import itertools
import copy
import json


def unitary_matrix(circuit, num_wires, *circ_args, **circ_kwargs):
    """Constructs the unitary matrix representation of a quantum
    circuit in the computational basis.

    :param circuit: A quantum function.
    :type circuit: Function

    :param num_wires: The number of wires needed by ``circuit``.
    :type num_wires: Int

    :param circ_args: Passthrough arguments for ``circuit``.
    :type circ_args: Positional Arguments

    :param circ_kwargs: Passthrough keyword arguments for ``circuit``.
    :type circ_kwargs: keyword Arguments

    :returns: A unitary matrix representing the provided ``circuit``.
    :rtype: Numpy Array
    """
    dev = qml.device("default.qubit", wires=range(num_wires))

    @qml.qnode(dev)
    def unitary_z(basis_state):
        qml.BasisState(basis_state, wires=range(num_wires))
        circuit(*circ_args, **circ_kwargs)
        return qml.state()

    bitstrings = list(itertools.product([0, 1], repeat=num_wires))
    u = [unitary_z(bitstring).numpy() for bitstring in bitstrings]
    return np.array(u).T


def write_optimization_json(opt_dict, filename):
    """Writes the optimization dictionary to a JSON file.

    :param opt_dict: The dictionary returned by a network optimization.
    :type opt_dict: dict

    :param filename: The name of the JSON file to be written. Note that ``.json`` extension is automatically added.
    :type filename: string

    :returns: ``None``
    """

    opt_dict_json = copy.deepcopy(opt_dict)

    opt_dict_json["opt_score"] = float(opt_dict_json["opt_score"])
    opt_dict_json["scores"] = [float(score) for score in opt_dict_json["scores"]]
    opt_dict_json["opt_settings"] = settings_to_list(opt_dict_json["opt_settings"])
    opt_dict_json["settings_history"] = [
        settings_to_list(settings) for settings in opt_dict_json["settings_history"]
    ]

    with open(filename + ".json", "w") as file:
        file.write(json.dumps(opt_dict_json, indent=2))


def read_optimization_json(filepath):
    """Reads data from an optimization JSON created via ``write_optimization_json``.

    :param filepath: The path to the JSON file. Note this string must contain the ``.json`` extension.
    :type filepath: string

    :returns: The optimization dictionary read from the file.
    :rtype: dict
    """

    with open(filepath) as file:
        opt_dict = json.load(file)

    opt_dict["opt_settings"] = settings_to_np(opt_dict["opt_settings"])
    opt_dict["settings_history"] = [
        settings_to_np(settings) for settings in opt_dict["settings_history"]
    ]

    return opt_dict


def settings_to_list(np_scenario_settings):
    """Converts the numpy array in scenario settings to lists.
    This function is intended for printing purposes.

    :param np_scenario_settings: The scenario settings structure for a `NetworkAnsatz`.

    :returns: The same nested array elements and structure using lists.
    """

    list_prep_settings = [node_settings.tolist() for node_settings in np_scenario_settings[0]]
    list_meas_settings = [node_settings.tolist() for node_settings in np_scenario_settings[1]]

    return [list_prep_settings, list_meas_settings]


def settings_to_np(list_scenario_settings):
    """Converts the nested list elements in the scenario settings to numpy arrays.
    This function is intended for printing purposes.

    :param list_scenario_settings: The scenario settings structure for a `NetworkAnsatz`.

    :returns: The same nested array elements and structure using numpy arrays.
    """

    np_prep_settings = [np.array(node_settings) for node_settings in list_scenario_settings[0]]
    np_meas_settings = [np.array(node_settings) for node_settings in list_scenario_settings[1]]

    return [np_prep_settings, np_meas_settings]


def mixed_base_num(n, base_digits):
    """Converts a base-10 number ``n`` into a mixed base number with digit
    values described by the ``base_digits`` array.

    :param n: A base-10 number
    :type n: int

    :param base_digits: A list of integers representing the largest value for each
                        digit in the mixed base number
    :type base_digits: list[int]

    :returns: A list of integers representing the mixed base number.
    :rtype: list[int]

    """
    mixed_base_digits = []
    n_tmp = n
    for i in range(len(base_digits)):
        place = int(math.prod(base_digits[i + 1 :]))

        mixed_base_digits += [n_tmp // place]
        n_tmp = n_tmp % place

    return mixed_base_digits



from qnetvo.utilities import *
from qnetvo.network_ansatz import *
from qnetvo.ansatz_library import *
from qnetvo.information import *
from qnetvo.cost.qnodes import *
from qnetvo.cost.postprocessing import *
from qnetvo.cost.I_3322_bell_inequality import *
from qnetvo.cost.mermin_klyshko_inequality import *
from qnetvo.cost.nlocal_chain_bell_inequality import *
from qnetvo.cost.nlocal_star_bell_inequality import *
from qnetvo.cost.magic_squares_game import *
from qnetvo.cost.chsh_inequality import *
from qnetvo.cost.linear_inequalities import *
from qnetvo.cost.mutual_info import *
from qnetvo.gradient_descent import *

# adding the quantum channels to "default.mixed" device
from pennylane.devices import DefaultMixed

DefaultMixed.operations.update(["two_qubit_depolarizing", "colored_noise"])



"""
Lazily imports dask when needed.
Throws an error if the user does not have dask[delayed] installed.
"""
try:
    import dask

    from dask import delayed as dask_delayed
    from dask import compute as dask_compute

except ImportError as e:  # pragma: no cover
    raise ImportError(
        "Dask must be installed for parallel evaluation. "
        "\nDask can be installed using pip:"
        "\n\npip install dask[delayed]"
    ) from e



from pennylane import math
from .cost.qnodes import joint_probs_qnode
from .utilities import mixed_base_num
from pennylane import numpy as np


def behavior_fn(network_ansatz, postmap=np.array([]), qnode_kwargs={}):
    """Creates an ansatz-specific function for constructing the behavior matrix.

    A behavior function is created as ``P_Net = behavior(network_ansatz)``
    and called as ``P_Net(ansatz_settings)``.
    The network behavior ``P_Net`` is a column stochastic matrix containing the
    conditional probabilities,

    .. math::

        \\mathbf{P}_{Net} = \\sum_{x,y,z}P(z|x,y)|z\\rangle\\langle x,y|,


    where :math:`P(z|x,y)` is evaluated by a qnode for each set of inputs
    :math:`x` and :math:`y`.

    A post-processing map :math:`\\mathbf{L}` may optionally be applied as
    :math:`\\mathbf{L}\\mathbf{P}_{Net}` where

    .. math::

        \\mathbf{L} = \\sum_{z',z}P(z'|z)|z'\\rangle\\langle z|.


    In the above expression, :math:`z'` is a new output drawn from a new alphabet.

    :param network_ansatz: A class describing the particular quantum network.
    :type network_ansatz: NetworkAnsatz

    :param postmap: A post-processing map applied to the bitstrings output from the
                     quantum circuit. The ``postmap`` matrix is column stochastic, that is,
                     each column sums to one and contains only positive values.
    :type postmap: *optional* np.ndarray

    :returns: A function ``P_Net(scenario_settings)`` that evaluates the
              behavior matrix for a given set of settings.
    :rtype: function
    """
    num_in_prep_nodes = [node.num_in for node in network_ansatz.prepare_nodes]
    num_in_meas_nodes = [node.num_in for node in network_ansatz.measure_nodes]

    num_out_meas_nodes = [node.num_out for node in network_ansatz.measure_nodes]

    base_digits = num_in_prep_nodes + num_in_meas_nodes
    net_num_in = math.prod(base_digits)
    net_num_out = math.prod(num_out_meas_nodes)

    raw_net_num_out = 2 ** len(network_ansatz.measure_wires)

    has_postmap = len(postmap) != 0
    if has_postmap:
        if postmap.shape[1] != raw_net_num_out:
            raise ValueError("The `postmap` must have " + str(raw_net_num_out) + " columns.")

    node_input_ids = [mixed_base_num(i, base_digits) for i in range(net_num_in)]

    probs_qnode = joint_probs_qnode(network_ansatz, **qnode_kwargs)

    def behavior(scenario_settings):
        raw_behavior = np.zeros((raw_net_num_out, net_num_in))
        for (i, input_id_set) in enumerate(node_input_ids):
            settings = network_ansatz.qnode_settings(
                scenario_settings,
                input_id_set[0 : len(num_in_prep_nodes)],
                input_id_set[len(num_in_prep_nodes) :],
            )

            raw_behavior[:, i] += probs_qnode(settings)

        return postmap @ raw_behavior if has_postmap else raw_behavior

    return behavior


def shannon_entropy(probs):
    """Evaluates the Shannon entropy for the given marginal probability
    distribution ``probs``.

    .. math::

        H(X) = -\\sum_{x\\in X}P(x)\\log_2(P(x))

    :param probs: A normalized probability vector.
    :type probs: np.array

    :returns: The Shannon entropy.
    :rtype: float
    """
    return -(
        math.sum([px * math.log2(px) if px != 0 and not (np.isclose(px, 0)) else 0 for px in probs])
    )



"""
Lazily imports tensorflow when needed.
Throws an error if the user does not have tensorflow installed.
"""
try:
    import tensorflow

    from tensorflow import keras

except ImportError as e:  # pragma: no cover
    raise ImportError(
        "TensorFlow is required for this functionality and is currently installed."
        "\nTensorFlow can be installed using pip:"
        "\n\npip install tensorflow"
    ) from e



import pennylane as qml
from pennylane.operation import Channel
from pennylane import numpy as np
import math

eps = 1e-7  # constant


def max_entangled_state(settings, wires):
    """Ansatz function for maximally entangled two-qubit state preparation.

    A general ``qml.Rot`` unitary is applied to one-side of a
    Bell state creating a general parameterization of all maximally
    entangled states.

    param settings: A list with three elements parameterizing a general single-qubit unitary
    :type settings: list[float]

    :param wires: The two wires on which the maximally entangled state is prepared.
    :type wires: qml.Wires
    """
    # prepare bell state
    qml.Hadamard(wires=wires[0])
    qml.CNOT(wires=wires[0:2])

    # perform general rotation on first qubit
    qml.Rot(*settings, wires=wires[0])


def bell_state_copies(settings, wires):
    """Initializes :math:`n` Bell states on :math:`2n` wires.
    The first :math:`n` wires represent Alice's half of the entangled
    state while Bob's half consists of the remaining :math:`n` wires.
    Entanglement is shared between Alice and Bob, however, the independent
    Bell states are not initially entangled with each other.

    :param settings: A placeholder parameter that is not used.
    :type settings: list[empty]

    :param wires: The wires on which the bell states are prepared.
    :type wires: qml.Wires
    """
    for i in range(len(wires) // 2):
        qml.Hadamard(wires=wires[i])
        qml.CNOT(wires=[wires[i], wires[i + len(wires) // 2]])


def ghz_state(settings, wires):
    """Initializes a GHZ state on the provided wires.

    :param settings: A placeholder parameter that is not used.
    :type settings: list[empty]

    :param wires: The wires on which the GHZ state is prepared.
    :type wires: qml.Wires
    """
    qml.Hadamard(wires=wires[0])

    for i in range(1, len(wires)):
        qml.CNOT(wires=[wires[0], wires[i]])


def local_RY(settings, wires):
    """Performs a rotation about :math:`y`-axis on each qubit
    specified by ``wires``.

    :param settings: A list of ``len(wires)`` real values.
    :type settings: list[float]

    :param wires: The wires on which the rotations are applied.
    :type wires: qml.Wires
    """
    qml.broadcast(qml.RY, wires, "single", settings)


def local_RXRY(settings, wires):
    """Performs a rotation about the :math:`x` and :math:`y` axes on
    each qubit specified by ``wires``.

    :param settings: A list of ``2*len(wires)`` real values.
    :type settings: list[float]

    :param wires: The wires on which the rotations are applied.
    :type wires: qml.Wires
    """
    for i, wire in enumerate(wires):
        qml.RX(settings[2 * i], wires=wire)
        qml.RY(settings[2 * i + 1], wires=wire)


def pure_amplitude_damping(noise_params, wires):
    """Implements a single qubit amplitude damping channel as a two-qubit gate.

    This allows amplitude damping noise to be applied on quantum hardware or on a
    state-vector simulator such as ``"default.qubit"``.

    This method is equivalent to the
    `pennylane.AmplitudeDamping <https://pennylane.readthedocs.io/en/stable/code/api/pennylane.AmplitudeDamping.html>`_
    method. The corresponding Kraus operators are expressed as:

    .. math::

        K_0 = \\begin{pmatrix}1 & 0 \\\\ 0 & \\sqrt{1-\\gamma} \\end{pmatrix}, \\quad
        K_1 = \\begin{pmatrix}0 & \\sqrt{\\gamma} \\\\ 0 & 0 \\end{pmatrix}.

    :param noise_params: The parameter ``[gamma]`` quantifying the amount of amplitude damping noise.
    :type noise_params: List[Float]

    :param wires: Two wires on which to implement the amplitude damping channel.
                  The channel input and output is on ``wires[0]`` while ``wires[1]``
                  serves as the ancilla register.
    :type wires: qml.Wires
    """

    ry_setting = 2 * math.asin(math.sqrt(noise_params[0]))

    qml.ctrl(qml.RY, control=wires[0])(ry_setting, wires=wires[1])
    qml.CNOT(wires=[wires[1], wires[0]])


def pure_phase_damping(noise_params, wires):
    """Implements a single qubit phase damping channel as a two-qubit gate.

    This allows phase damping noise to be applied on quantum hardware or on a
    state-vector simulator such as ``"default.qubit"``.

    This method is equivalent to the
    `pennylane.PhaseDamping <https://pennylane.readthedocs.io/en/stable/code/api/pennylane.PhaseDamping.html>`_
    method. The corresponding Kraus operators are expressed as:

    .. math::

        K_0 = \\begin{pmatrix}1 & 0 \\\\ 0 & \\sqrt{1-\\gamma} \\end{pmatrix}, \\quad
        K_1 = \\begin{pmatrix}0 & 0 \\\\ 0 & \\sqrt{\\gamma} \\end{pmatrix}.

    :param noise_params: The parameter ``[gamma]`` quantifying the amount of phase damping noise.
    :type noise_params: List[Float]

    :param wires: Two wires on which to implement the phase damping channel.
                  The channel input and output is on ``wires[0]`` while ``wires[1]``
                  serves as the ancilla register.
    :type wires: qml.Wires
    """

    ry_setting = 2 * math.asin(math.sqrt(noise_params[0]))
    qml.ctrl(qml.RY, control=wires[0])(ry_setting, wires=wires[1])


class two_qubit_depolarizing(Channel):
    """Applies a two-qubit depolarizing channel using Kraus operators.

    The channel is called as a quantum function ``two_qubit_depolarizing(gamma, wires)``

    :param gamma: The amount of depolarizing noise in the channel.
    :type gamma: Float

    :param wires: Two wires on which to apply the depolarizing noise.
    :type wires: qml.Wires

    For a noise parameter :math:`\\gamma`, the two-qubit depolarizing
    noise model is represented by the following function:

    .. math::

        \\mathcal{N}(\\rho) = (1-\\frac{16}{15}\\gamma)\\rho + (\\frac{16}{15})(\\frac{\\gamma}{4})) \\mathbb{I}

    :raises ValueError: if ``0 <= gamma <= 1`` is not satisfied.
    """

    num_params = 1
    num_wires = 2
    grad_method = "F"

    @classmethod
    def _kraus_matrices(cls, *params):
        gamma = params[0]

        if not 0.0 - eps <= gamma <= 1.0 + eps:
            raise ValueError("gamma must be in the interval [0,1].")
        elif np.isclose(gamma, 0):
            gamma = 0

        paulis = [
            np.array([[1, 0], [0, 1]]),
            np.array([[0, 1], [1, 0]]),
            np.array([[0, -1j], [1j, 0]]),
            np.array([[1, 0], [0, -1]]),
        ]

        kraus_ops = []
        for i in range(4):
            for j in range(4):
                scalar = np.sqrt(1 - gamma) if i + j == 0 else np.sqrt(gamma / 15)
                kraus_ops.append(scalar * np.kron(paulis[i], paulis[j]))

        return kraus_ops


class colored_noise(Channel):
    """Applies a two-qubit colored noise channel using Kraus operators.

    The channel is called as a quantum function ``colored_noise(gamma, wires)``.

    :param gamma: The amount of colored noise in the channel.
    :type gamma: Float

    :param wires: Two wires on which to apply the colored noise.
    :type wires: qml.Wires

    For the noise parameter :math:`\\gamma`, the colored noise model is
    represented by the following function:

    .. math::

        \\mathcal{N}(\\rho) = (1-\\gamma)\\rho + \\frac{\\gamma}{2}(|01\\rangle\\langle 01| + |10\\rangle\\langle 10|)

    :raises ValueError: if ``0 <= gamma <= 1`` is not satisfied.
    """

    num_params = 1
    num_wires = 2
    grad_method = "F"

    @classmethod
    def _kraus_matrices(cls, *params):
        gamma = params[0]

        if not 0.0 - eps <= gamma <= 1.0 + eps:
            raise ValueError("gamma must be in the interval [0,1].")
        elif np.isclose(gamma, 0):
            gamma = 0

        phi_plus = np.array([1, 0, 0, 1]) / np.sqrt(2)
        phi_min = np.array([1, 0, 0, -1]) / np.sqrt(2)
        psi_plus = np.array([0, 1, 1, 0]) / np.sqrt(2)
        psi_min = np.array([0, 1, -1, 0]) / np.sqrt(2)

        scalar = np.sqrt(gamma / 2)

        K0 = np.sqrt(1 - gamma) * np.eye(4)
        K1 = scalar * np.outer(psi_plus, phi_plus)
        K2 = scalar * np.outer(psi_plus, phi_min)
        K3 = scalar * np.outer(psi_plus, psi_plus)
        K4 = scalar * np.outer(psi_plus, psi_min)
        K5 = scalar * np.outer(psi_min, phi_plus)
        K6 = scalar * np.outer(psi_min, phi_min)
        K7 = scalar * np.outer(psi_min, psi_plus)
        K8 = scalar * np.outer(psi_min, psi_min)
        return [K0, K1, K2, K3, K4, K5, K6, K7, K8]
